@@ -1,4 +1,4 @@
-// File created from ../../../src/lib/http/auth_messages.mes
+// File created from src/lib/http/auth_messages.mes
 
 #ifndef AUTH_MESSAGES_H
 #define AUTH_MESSAGES_H
@@ -8,10 +8,12 @@
 namespace isc {
 namespace http {
 
+extern const isc::log::MessageID HTTP_CLIENT_PASSWORD_SECURITY_WARNING;
 extern const isc::log::MessageID HTTP_CLIENT_REQUEST_AUTHORIZED;
 extern const isc::log::MessageID HTTP_CLIENT_REQUEST_BAD_AUTH_HEADER;
 extern const isc::log::MessageID HTTP_CLIENT_REQUEST_NOT_AUTHORIZED;
 extern const isc::log::MessageID HTTP_CLIENT_REQUEST_NO_AUTH_HEADER;
+extern const isc::log::MessageID HTTP_CLIENT_USER_SECURITY_WARNING;
 
 } // namespace http
 } // namespace isc
